# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
"""
Config example for this dag:
{
    "datarobot_jdbc_connection": "datarobot_jdbc_test",
    "dataset_name": "test_dataset_name",
    "table_schema": "integration_demo",
    "table_name": "test_table",
    "query": 'SELECT * FROM "integration_demo"."test_table"',
    "persist_data_after_ingestion": True,
    "do_snapshot": True,
}
"""

from datetime import datetime

from airflow.decorators import dag

from datarobot_provider.operators.ai_catalog import CreateDatasetFromDataStoreOperator
from datarobot_provider.operators.datarobot import CreateProjectOperator


@dag(
    schedule=None,
    start_date=datetime(2023, 1, 1),
    tags=["example"],
    # Default json config example:
    params={
        "project_name": "test_project_name",
        "datarobot_jdbc_connection": "datarobot_jdbc_test_connection_name",
        "dataset_name": "test_jdbc_dataset_name",
        "table_schema": "test_jdbc_table_schema",
        "table_name": "test_jdbc_table_name",
        "persist_data_after_ingestion": True,
        "do_snapshot": True,
    },
)
def datarobot_dataset_connect():
    dataset_connect_op = CreateDatasetFromDataStoreOperator(
        task_id="create_dataset_jdbc",
    )

    create_project_op = CreateProjectOperator(
        task_id="create_project",
        dataset_id=dataset_connect_op.output,
    )

    dataset_connect_op >> create_project_op


datarobot_jdbc_connection_dag = datarobot_dataset_connect()

if __name__ == "__main__":
    datarobot_jdbc_connection_dag.test()



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from datetime import datetime

from airflow.decorators import dag
from datarobot import AUTOPILOT_MODE

from datarobot_provider.operators.autopilot import StartAutopilotOperator
from datarobot_provider.operators.datarobot import CreateProjectOperator
from datarobot_provider.operators.feature_discovery import DatasetDefinitionOperator
from datarobot_provider.operators.feature_discovery import DatasetRelationshipOperator
from datarobot_provider.operators.feature_discovery import RelationshipsConfigurationOperator
from datarobot_provider.sensors.datarobot import AutopilotCompleteSensor


@dag(
    schedule=None,
    start_date=datetime(2022, 1, 1),
    tags=["example", "feature discovery"],
    params={
        "project_name": "Demo Airflow DataRobot Feature Discovery",
        "autopilot_settings": {
            "target": "BadLoan",
            "mode": AUTOPILOT_MODE.QUICK,
            "feature_engineering_prediction_point": "date",
        },
        "advanced_options": {"feature_discovery_supervised_feature_reduction": True},
        "feature_discovery_settings": [
            {"name": "enable_days_from_prediction_point", "value": True},
            {"name": "enable_hour", "value": True},
            {"name": "enable_categorical_num_unique", "value": False},
            {"name": "enable_categorical_statistics", "value": False},
            {"name": "enable_numeric_minimum", "value": True},
            {"name": "enable_token_counts", "value": False},
            {"name": "enable_latest_value", "value": True},
            {"name": "enable_numeric_standard_deviation", "value": True},
            {"name": "enable_numeric_skewness", "value": False},
            {"name": "enable_day_of_week", "value": True},
            {"name": "enable_entropy", "value": False},
            {"name": "enable_numeric_median", "value": True},
            {"name": "enable_word_count", "value": False},
            {"name": "enable_pairwise_time_difference", "value": True},
            {"name": "enable_days_since_previous_event", "value": True},
            {"name": "enable_numeric_maximum", "value": True},
            {"name": "enable_numeric_kurtosis", "value": False},
            {"name": "enable_most_frequent", "value": False},
            {"name": "enable_day", "value": True},
            {"name": "enable_numeric_average", "value": True},
            {"name": "enable_summarized_counts", "value": False},
            {"name": "enable_missing_count", "value": True},
            {"name": "enable_record_count", "value": True},
            {"name": "enable_numeric_sum", "value": True},
        ],
        "unsupervised_mode": False,
        "use_feature_discovery": True,
    },
)
def datarobot_feature_discovery_pipeline(
    training_dataset_id="64ea43b4465047fedb5ab3b2",
    training_dataset_version_id="64ea43b5465047fedb5ab3b3",
    profile_dataset_identifier="profile",
    profile_dataset_id="64ea423b666704a28e8fd613",
    profile_dataset_version_id="64ea423c666704a28e8fd614",
    transaction_dataset_identifier="transactions",
    transaction_dataset_id="64ea42e1666704a28e8fd6ac",
    transaction_dataset_version_id="64ea42e1666704a28e8fd6ad",
):
    create_project_op = CreateProjectOperator(
        task_id="create_project",
        dataset_id=training_dataset_id,
        dataset_version_id=training_dataset_version_id,
    )

    profile_dataset_definition_op = DatasetDefinitionOperator(
        task_id="profile_dataset_definition",
        dataset_identifier=profile_dataset_identifier,
        dataset_id=profile_dataset_id,
        dataset_version_id=profile_dataset_version_id,
    )

    transaction_dataset_definition_op = DatasetDefinitionOperator(
        task_id="transaction_dataset_definition",
        dataset_identifier=transaction_dataset_identifier,
        dataset_id=transaction_dataset_id,
        dataset_version_id=transaction_dataset_version_id,
        primary_temporal_key="Date",
    )

    # You do not need to specify dataset1Identifier when joining with the primary dataset
    primary_profile_relationship_op = DatasetRelationshipOperator(
        task_id="primary_profile_relationship",
        dataset2_identifier=profile_dataset_identifier,  # to profile
        dataset1_keys=["CustomerID"],  # on CustomerID
        dataset2_keys=["CustomerID"],
        feature_derivation_windows=[
            {"start": -7, "end": -1, "unit": "DAY"},
            {"start": -14, "end": -1, "unit": "DAY"},
            {"start": -30, "end": -1, "unit": "DAY"},
        ],  # example of multiple FDW
        prediction_point_rounding=1,
        prediction_point_rounding_time_unit="DAY",
    )

    relationship_definition_op = DatasetRelationshipOperator(
        task_id="relationship_definition",
        dataset1_identifier=profile_dataset_identifier,  # join profile
        dataset2_identifier=transaction_dataset_identifier,  # to transactions
        dataset1_keys=["CustomerID"],  # on CustomerID
        dataset2_keys=["CustomerID"],
    )

    relationships_configuration_op = RelationshipsConfigurationOperator(
        task_id="relationships_configuration",
        dataset_definitions=[
            profile_dataset_definition_op.output,
            transaction_dataset_definition_op.output,
        ],
        relationships=[primary_profile_relationship_op.output, relationship_definition_op.output],
    )

    train_models_op = StartAutopilotOperator(
        task_id="train_models",
        project_id=create_project_op.output,
        relationships_configuration_id=relationships_configuration_op.output,
    )

    autopilot_complete_sensor = AutopilotCompleteSensor(
        task_id="check_autopilot_complete",
        project_id=create_project_op.output,
    )

    (
        (profile_dataset_definition_op, transaction_dataset_definition_op)
        >> create_project_op
        >> (primary_profile_relationship_op, relationships_configuration_op)
        >> train_models_op
        >> autopilot_complete_sensor
    )


datarobot_feature_discovery_pipeline_dag = datarobot_feature_discovery_pipeline()

if __name__ == "__main__":
    datarobot_feature_discovery_pipeline_dag.test()



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.

from datetime import datetime

from airflow.decorators import dag
from airflow.operators.empty import EmptyOperator
from airflow.operators.python import BranchPythonOperator
from datarobot import TARGET_TYPE
from datarobot.enums import MODEL_REPLACEMENT_REASON

from datarobot_provider.operators.custom_models import CreateCustomModelVersionOperator
from datarobot_provider.operators.custom_models import CustomModelTestOperator
from datarobot_provider.operators.custom_models import GetCustomModelTestOverallStatusOperator
from datarobot_provider.operators.deployment import ReplaceModelOperator


@dag(
    schedule=None,
    start_date=datetime(2023, 1, 1),
    tags=["example", "custom model"],
    params={
        "programming_language": "python",
        "required_metadata_keys": [],
        "custom_model_description": "This is a custom model created by Airflow",
        "environment_version_description": "created by Airflow provider",
        "custom_model_name": "Airflow Custom model Demo",
        "target_type": TARGET_TYPE.REGRESSION,
        "target_name": "Grade 2014",
        "is_major_update": True,
        "is_training_data_for_versions_permanently_enabled": True,
        "docker_context_path": "/usr/local/airflow/dags/datarobot-user-models/public_dropin_environments/python3_pytorch/",
        "custom_model_folder": "/usr/local/airflow/dags/datarobot-user-models/model_templates/python3_pytorch/",
        "test_dataset_file_path": "/usr/local/airflow/dags/datarobot-user-models/tests/testdata/juniors_3_year_stats_regression.csv",
        "train_dataset_file_path": "/usr/local/airflow/dags/datarobot-user-models/tests/testdata/juniors_3_year_stats_regression.csv",
    },
)
def update_custom_model_pipeline(
    execution_environment_id="650d8b27e53f3f2bb067b58a",
    custom_inference_model_id="650d8db8473c77acc713d239",
    train_dataset_id="65106d23d7fe593f61995bad",
    test_dataset_id="65106d23d7fe593f61995bad",
    deployment_id="650d90d1c03bf7a69afdda85",
):
    create_custom_model_version_op = CreateCustomModelVersionOperator(
        task_id="create_custom_model_version",
        custom_model_id=custom_inference_model_id,
        base_environment_id=execution_environment_id,
        training_dataset_id=train_dataset_id,
    )

    custom_model_test_op = CustomModelTestOperator(
        task_id="custom_model_test",
        custom_model_id=custom_inference_model_id,
        custom_model_version_id=create_custom_model_version_op.output,
        dataset_id=test_dataset_id,
    )

    custom_model_test_overall_status_op = GetCustomModelTestOverallStatusOperator(
        task_id="custom_model_test_overall_status",
        custom_model_test_id=custom_model_test_op.output,
    )

    def choose_branch(custom_model_test_overall_status):
        if custom_model_test_overall_status == "succeeded":
            return ["replace_deployment_model"]
        else:
            return ["model_replacement_fail"]

    if_tests_succeed = BranchPythonOperator(
        task_id="if_tests_succeed",
        python_callable=choose_branch,
        op_args=[
            custom_model_test_overall_status_op.output,
        ],
    )

    model_replacement_fail = EmptyOperator(
        task_id="model_replacement_fail",
    )

    replace_deployment_model_op = ReplaceModelOperator(
        task_id="replace_deployment_model",
        deployment_id=deployment_id,
        new_model_id=create_custom_model_version_op.output,
        reason=MODEL_REPLACEMENT_REASON.ACCURACY,
    )

    (
        create_custom_model_version_op
        >> custom_model_test_op
        >> custom_model_test_overall_status_op
        >> if_tests_succeed
        >> [
            replace_deployment_model_op,
            model_replacement_fail,
        ]
    )


update_custom_model_pipeline_dag = update_custom_model_pipeline()

if __name__ == "__main__":
    update_custom_model_pipeline_dag.test()



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from collections.abc import Sequence
from typing import Any

import datarobot as dr
from airflow.exceptions import AirflowException
from airflow.exceptions import AirflowNotFoundException
from airflow.utils.context import Context

from datarobot_provider.hooks.connections import JDBCDataSourceHook
from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator


class GetOrCreateDataStoreOperator(BaseDatarobotOperator):
    """
    !! Please, manage database connections via DataRobot rather than Airflow and use *GetDataStoreOperator* instead. !!

    Fetching DataStore by connection name or creating if it does not exist
    and return DataStore ID.

    :param connection_param_name: name of parameter in the config file corresponding to connection name
    :type connection_param_name: str, optional
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: DataRobot Credentials ID
    :rtype: str
    """

    def __init__(
        self,
        *,
        connection_param_name: str = "datarobot_connection_name",
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.connection_param_name = connection_param_name

    def execute(self, context: Context) -> str:
        if self.connection_param_name not in context["params"]:
            raise AirflowNotFoundException(
                f"Attribute: {self.connection_param_name} not present in config"
            )
        # Getting connection name from config:
        connection_name = context["params"][self.connection_param_name]

        # Fetch stored JDBC Connection with credentials
        _, _, data_store = JDBCDataSourceHook(datarobot_credentials_conn_id=connection_name).run()

        if data_store is not None:
            self.log.info(f"Found preconfigured jdbc connection: {connection_name}")

        return data_store.id


class GetDataStoreOperator(BaseDatarobotOperator):
    """Get a DataRobot data store id by data connection name.
    You have to create a DataRobot data connection in advance at /account/data-connections page.

    :param data_connection: unique, case-sensitive data connection name as you can see it at DataRobot.
    :type data_connection: str
    :return: Data store ID.
    :rtype: str
    """

    template_fields: Sequence[str] = ["data_connection"]

    def __init__(
        self,
        *,
        data_connection: str = "{{ params.data_connection }}",
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.data_connection = data_connection

    def execute(self, context: Context) -> Any:
        for datastore in dr.DataStore.list(name=self.data_connection):
            if datastore.canonical_name == self.data_connection:
                break

        else:
            raise AirflowException(f"Connection {self.data_connection} was not found.")

        return datastore.id



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import TYPE_CHECKING
from typing import Any
from typing import Optional
from typing import Sequence

import datarobot as dr
from airflow.exceptions import AirflowFailException
from airflow.utils.context import Context
from datarobot.enums import MODEL_REPLACEMENT_REASON

from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator
from datarobot_provider.operators.datarobot import DATAROBOT_MAX_WAIT
from datarobot_provider.operators.datarobot import DATETIME_FORMAT

if TYPE_CHECKING:
    from datarobot.models.deployment.deployment import ModelDict


def _serialize_drift(drift_obj, date_format=DATETIME_FORMAT):
    drift_dict = drift_obj.__dict__.copy()
    drift_dict["period"] = {
        "start": drift_obj.period["start"].strftime(date_format),
        "end": drift_obj.period["end"].strftime(date_format),
    }
    return drift_dict


class DeployModelMixin:
    def deploy_model(
        self, model_id: str, label: str, description: Optional[str] = None
    ) -> dr.Deployment:
        """Deploys the provided model to production."""
        self.log.info(f"Deploying model_id={model_id} with label={label}")  # type: ignore[attr-defined]
        prediction_server = dr.PredictionServer.list()[0]
        deployment = dr.Deployment.create_from_learning_model(
            model_id, label, description, prediction_server.id
        )
        self.log.info(f"Model deployed: deployment_id={deployment.id}")  # type: ignore[attr-defined]
        self.log.info("Enabling tracking for target drift and feature drift")  # type: ignore[attr-defined]
        deployment.update_drift_tracking_settings(
            target_drift_enabled=True, feature_drift_enabled=True, max_wait=DATAROBOT_MAX_WAIT
        )
        return deployment


class DeployModelOperator(BaseDatarobotOperator, DeployModelMixin):
    """
    Deploys the specified model to production.

    :param model_id: ID of the DataRobot model to deploy
    :type model_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: DataRobot deployment ID
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["model_id"]

    def __init__(
        self,
        *,
        model_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.model_id = model_id

    def execute(self, context: Context) -> str:
        # Deploy the model
        deployment = self.deploy_model(
            self.model_id,
            context["params"]["deployment_label"],
            context["params"].get("deployment_description"),
        )
        return deployment.id


class DeployRecommendedModelOperator(BaseDatarobotOperator, DeployModelMixin):
    """
    Deploys a recommended model to production.

    :param project_id: DataRobot project ID
    :type project_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: DataRobot deployment ID
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["project_id"]

    def __init__(
        self,
        *,
        project_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id

    def deploy_recommended_model(
        self, project_id: str, label: str, description: Optional[str] = None
    ) -> dr.Deployment:
        """Deploys the recommended model to production."""
        self.log.info(f"Retrieving recommended model for project_id={project_id}")
        project: dr.Project = dr.Project.get(project_id)
        model = project.recommended_model()
        if model is None:
            raise AirflowFailException(f"No recommended model found for project_id={project_id}")
        return self.deploy_model(model.id, label, description)

    def execute(self, context: Context) -> str:
        # Deploy the recommended model
        deployment = self.deploy_recommended_model(
            self.project_id,
            context["params"]["deployment_label"],
            context["params"].get("deployment_description"),
        )
        return deployment.id


class GetDeploymentModelOperator(BaseDatarobotOperator):
    """
    Gets current model info from a deployment.

    :param deployment_id: DataRobot deployment ID
    :type deployment_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: model info from a Deployment
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["deployment_id"]

    def __init__(
        self,
        *,
        deployment_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_id = deployment_id

    def execute(self, context: Context) -> Optional["ModelDict"]:
        self.log.info(f"Getting model_id for deployment_id={self.deployment_id}")
        deployment = dr.Deployment.get(self.deployment_id)
        return deployment.model


class GetDeploymentStatusOperator(BaseDatarobotOperator):
    """
    Get a Deployment status (active/inactive).

    :param deployment_id: DataRobot deployment ID
    :type deployment_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: Deployment status (active/inactive)
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating

    template_fields: Sequence[str] = ["deployment_id"]

    def __init__(
        self,
        *,
        deployment_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_id = deployment_id

    def validate(self) -> None:
        if not self.deployment_id:
            raise ValueError("Invalid or missing `deployment_id` value")

    def execute(self, context: Context) -> Optional[str]:
        self.log.info(f"Getting Deployment for deployment_id={self.deployment_id}")
        deployment = dr.Deployment.get(self.deployment_id)
        return deployment.status


class ReplaceModelOperator(BaseDatarobotOperator):
    """
    Replaces the current model for a deployment.

    :param deployment_id: DataRobot deployment ID
    :type deployment_id: str
    :param new_model_id: The id of the new model to use. If replacing the deployment's model with a
            CustomInferenceModel, a specific CustomModelVersion ID must be used.
    :type new_model_id: str
    :param reason: str
            The reason for the model replacement. Must be one of 'ACCURACY', 'DATA_DRIFT', 'ERRORS',
            'SCHEDULED_REFRESH', 'SCORING_SPEED', or 'OTHER'. This value will be stored in the model
            history to keep track of why a model was replaced
    :type reason: str, optional
    :param max_wait_sec: The maximum time to wait for
            model replacement job to complete before erroring,
            defaults to 600 seconds
    :type max_wait_sec: int, optional
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["deployment_id", "new_model_id", "reason"]

    def __init__(
        self,
        *,
        deployment_id: str,
        new_model_id: str,
        reason: str = MODEL_REPLACEMENT_REASON.OTHER,
        max_wait_sec: int = 600,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_id = deployment_id
        self.new_model_id = new_model_id
        self.reason = reason
        self.max_wait_sec = max_wait_sec

    def validate(self) -> None:
        if self.deployment_id is None:
            raise AirflowFailException("deployment_id must be provided")

        if self.new_model_id is None:
            raise AirflowFailException("new_model_id must be provided")

    def execute(self, context: Context) -> None:
        self.log.info(f"Getting model_id for deployment_id={self.deployment_id}")
        deployment = dr.Deployment.get(deployment_id=self.deployment_id)
        self.log.info(
            f"Validating replacement model new_model_id={self.new_model_id} for deployment_id={self.deployment_id}"
        )
        check_result, check_message, status_list = deployment.validate_replacement_model(
            new_model_id=self.new_model_id
        )
        self.log.info(f"Validation result: {check_result}, message: {check_message}")
        self.log.info(f"Validation result details: {status_list}")
        if check_result == "failing":
            raise AirflowFailException(check_message)
        self.log.info(
            f"Trying to replace a model for deployment_id={self.deployment_id} to new_model_id={self.new_model_id}"
        )
        deployment.replace_model(
            new_model_id=self.new_model_id, reason=self.reason, max_wait=self.max_wait_sec
        )
        self.log.info(
            f"Model for deployment_id={self.deployment_id} replaced to new_model_id={self.new_model_id}"
        )


class ScorePredictionsOperator(BaseDatarobotOperator):
    """
    Creates a batch prediction job in DataRobot, scores the data and saves prediction to the output.
    :param deployment_id: DataRobot deployment ID
    :type deployment_id: str
    :param intake_datastore_id: DataRobot DataStore ID for jdbc source connection
    :type intake_datastore_id: str
    :param intake_credential_id: DataRobot Credentials ID for source connection
    :type intake_credential_id: str
    :param output_datastore_id: DataRobot DataStore ID for jdbc destination connection
    :type output_datastore_id: str
    :param output_credential_id: DataRobot Credentials ID for destination connection
    :type output_credential_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: Batch predictions job ID
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "deployment_id",
        "intake_datastore_id",
        "intake_credential_id",
        "output_datastore_id",
        "output_credential_id",
        "score_settings",
    ]

    def __init__(
        self,
        *,
        deployment_id: Optional[str] = None,
        intake_datastore_id: Optional[str] = None,
        intake_credential_id: Optional[str] = None,
        output_datastore_id: Optional[str] = None,
        output_credential_id: Optional[str] = None,
        score_settings: Optional[dict] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_id = deployment_id
        self.intake_datastore_id = intake_datastore_id
        self.intake_credential_id = intake_credential_id
        self.output_datastore_id = output_datastore_id
        self.output_credential_id = output_credential_id
        self.score_settings = score_settings

    def execute(self, context: Context) -> str:
        if self.score_settings is None:
            self.score_settings = context["params"]["score_settings"]

        # in case of deployment_id was not set from operator argument:
        if self.deployment_id is None:
            self.deployment_id = self.score_settings["deployment_id"]

        if self.intake_credential_id is not None:
            self.score_settings["intake_settings"]["credential_id"] = self.intake_credential_id
        if self.output_credential_id is not None:
            self.score_settings["output_settings"]["credential_id"] = self.output_credential_id

        intake_settings = self.score_settings.get("intake_settings", dict())
        output_settings = self.score_settings.get("output_settings", dict())

        intake_type = intake_settings.get("type")
        output_type = output_settings.get("type")

        # in case of JDBC intake from operator argument:
        if intake_type == "jdbc" and self.intake_datastore_id is not None:
            self.score_settings["intake_settings"]["data_store_id"] = self.intake_datastore_id

        # in case of JDBC output from operator argument:
        if output_type == "jdbc" and self.output_datastore_id is not None:
            self.score_settings["output_settings"]["data_store_id"] = self.output_datastore_id

        # Score data
        self.log.info(
            f"Scoring predictions against deployment_id={self.deployment_id} "
            f"with settings: {self.score_settings}"
        )

        # BatchPredictionJob.score() method in the Python SDK expects a DataRobot dataset instance
        if intake_type == "dataset":
            dataset_id = intake_settings.get("dataset_id")
            if not dataset_id:
                raise ValueError(
                    "Invalid or missing `dataset_id` value for the `dataset` intake type."
                )
            dataset = dr.Dataset.get(dataset_id)
            intake_settings["dataset"] = dataset

            # We no longer need the ID
            del intake_settings["dataset_id"]

        job = dr.BatchPredictionJob.score(self.deployment_id, **self.score_settings)
        self.log.info(f"Batch Prediction submitted, job.id={job.id}")
        return job.id


class ActivateDeploymentOperator(BaseDatarobotOperator):
    """
    Activate or deactivate a Deployment.

    :param deployment_id: DataRobot deployment ID
    :type deployment_id: str
    :param activate: if set to True - activate deployment, if set to False - deactivate deployment
    :type activate: boolean
    :param max_wait_sec: The maximum time in seconds to wait for deployment activation/deactivation to complete
    :type max_wait_sec: int
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: Deployment status (active/inactive)
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["deployment_id", "activate"]

    def __init__(
        self,
        *,
        deployment_id: str,
        activate: bool = True,
        max_wait_sec: int = 600,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_id = deployment_id
        self.activate = activate
        self.max_wait_sec = max_wait_sec

    def validate(self) -> None:
        if self.deployment_id is None:
            raise ValueError("Invalid or missing `deployment_id` value")

    def execute(self, context: Context) -> Optional[str]:
        self.log.info(f"Getting Deployment with deployment_id={self.deployment_id}")
        deployment = dr.Deployment.get(self.deployment_id)
        if self.activate:
            deployment.activate(max_wait=self.max_wait_sec)
        else:
            deployment.deactivate(max_wait=self.max_wait_sec)
        return deployment.status


class GetFeatureDriftOperator(BaseDatarobotOperator):
    """
    Gets feature drift measurements from a deployment.

    :param deployment_id: DataRobot deployment ID
    :type deployment_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: Drift stats for a Deployment's features
    :rtype: list[dict]
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["deployment_id"]

    def __init__(
        self,
        *,
        deployment_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_id = deployment_id

    def execute(self, context: Context) -> list[dict]:
        self.log.info(f"Getting feature drift for deployment_id={self.deployment_id}")
        deployment = dr.Deployment.get(self.deployment_id)
        feature_drift_params = context["params"].get("feature_drift", {})
        drift = deployment.get_feature_drift(**feature_drift_params)
        return [_serialize_drift(feature) for feature in drift]


class GetTargetDriftOperator(BaseDatarobotOperator):
    """
    Gets target drift measurements from a deployment.

    :param deployment_id: DataRobot deployment ID
    :type deployment_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: Drift stats for a Deployment's target
    :rtype: dict
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["deployment_id"]

    def __init__(
        self,
        *,
        deployment_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_id = deployment_id

    def execute(self, context: Context) -> dict[str, Any]:
        self.log.info(f"Getting target drift for deployment_id={self.deployment_id}")
        deployment = dr.Deployment.get(self.deployment_id)
        target_drift_params = context["params"].get("target_drift", {})
        drift = deployment.get_target_drift(**target_drift_params)
        return _serialize_drift(drift)



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from collections.abc import Sequence
from typing import Any
from typing import Iterable
from typing import Optional

import datarobot as dr
from airflow.utils.context import Context
from datarobot import SNAPSHOT_POLICY

from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator
from datarobot_provider.operators.base_datarobot_operator import BaseUseCaseEntityOperator

DATAROBOT_MAX_WAIT = 600


class CreateFeatureDiscoveryRecipeOperator(BaseUseCaseEntityOperator):
    """
    Create a Feature Discovery recipe.

    :param dataset_id: DataRobot Dataset Catalog ID to use as the primary dataset
    :type dataset_id: str
    :param use_case_id: ID of the use case to add the recipe to
    :type use_case_id: str
    :param dataset_definitions: list of dataset definitions
        Each element is a dict retrieved from DatasetDefinitionOperator operator
    :type dataset_definitions: Iterable[dict]
    :param relationships: list of relationships
        Each element is a dict retrieved from DatasetRelationshipOperator operator
    :type relationships: Iterable[dict]
    :param feature_discovery_settings: list of feature discovery settings, optional
        If not provided, it will be retrieved from DAG configuration params otherwise default
        settings will be used.
    :type feature_discovery_settings: Iterable[dict]
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: DataRobot Recipe ID
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "dataset_id",
        "use_case_id",
        "dataset_definitions",
        "relationships",
        "feature_discovery_settings",
    ]

    def __init__(
        self,
        *,
        dataset_id: str,
        dataset_definitions: Iterable[dict],
        relationships: Iterable[dict],
        feature_discovery_settings: Optional[Iterable[dict]] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.dataset_id = dataset_id
        self.dataset_definitions = dataset_definitions
        self.relationships = relationships
        self.feature_discovery_settings = feature_discovery_settings

    def execute(self, context: Context) -> str:
        use_case: dr.UseCase = self.get_use_case(context, required=True)  # type: ignore[assignment]
        dataset = dr.Dataset.get(self.dataset_id)

        recipe = dr.models.Recipe.from_dataset(
            use_case,
            dataset,
            recipe_type=dr.enums.RecipeType.FEATURE_DISCOVERY,
        )

        # Get dataset version ID if it isn't defined by the user:
        for dataset_definition in self.dataset_definitions:
            if not dataset_definition.get("catalogVersionId"):
                secondary_dataset = dr.Dataset.get(dataset_definition["catalogId"])
                dataset_definition["catalogVersionId"] = secondary_dataset.version_id

        recipe_id = recipe.id
        recipe_config_id = recipe.settings.relationships_configuration_id  # type: ignore[union-attr]

        # Add secondary dataset configuration information into the Recipe config
        dr.RelationshipsConfiguration(recipe_config_id).replace(
            self.dataset_definitions,
            self.relationships,
            self.feature_discovery_settings,
        )

        self.log.info(f"Feature Discovery recipe created: recipe_id={recipe_id}")

        return recipe_id


class RelationshipsConfigurationOperator(BaseDatarobotOperator):
    """
    Create a Relationships Configuration.

    :param dataset_definitions: list of dataset definitions
        Each element is a dict retrieved from DatasetDefinitionOperator operator
    :type dataset_definitions: Iterable[dict]
    :param relationships: list of relationships
        Each element is a dict retrieved from DatasetRelationshipOperator operator
    :type relationships: Iterable[dict]
    :param feature_discovery_settings: list of feature discovery settings, optional
        If not provided, it will be retrieved from DAG configuration params otherwise default
        settings will be used.
    :type feature_discovery_settings: Iterable[dict]
    :param max_wait_sec: For some settings, an asynchronous task must be run to analyze the
            dataset.  max_wait governs the maximum time (in seconds) to wait before giving up.
    :type max_wait_sec: int, optional
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: DataRobot Relationships Configuration ID
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "dataset_definitions",
        "relationships",
        "feature_discovery_settings",
    ]
    template_fields_renderers: dict[str, str] = {}
    template_ext: Sequence[str] = ()
    ui_color = "#f4a460"

    def __init__(
        self,
        *,
        dataset_definitions: Iterable[dict],
        relationships: Iterable[dict],
        feature_discovery_settings: Optional[Iterable[dict]] = None,
        max_wait_sec: int = DATAROBOT_MAX_WAIT,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.dataset_definitions = dataset_definitions
        self.relationships = relationships
        self.feature_discovery_settings = feature_discovery_settings
        self.max_wait_sec = max_wait_sec

    def execute(self, context: Context) -> None:
        # if feature_discovery_settings not provided, try to get it from DAG configuration params:
        if self.feature_discovery_settings is None:
            self.feature_discovery_settings = context["params"].get(
                "feature_discovery_settings", None
            )

        # Create the relationships' configuration to define the connection between the datasets
        # This will be passed to the DataRobot project configuration for Autopilot
        relationship_config = dr.RelationshipsConfiguration.create(
            dataset_definitions=self.dataset_definitions,
            relationships=self.relationships,
            feature_discovery_settings=self.feature_discovery_settings,
        )

        return relationship_config.id


class DatasetDefinitionOperator(BaseDatarobotOperator):
    """
    Dataset definition for the Feature Discovery

    :param dataset_identifier: Alias of the dataset (used directly as part of the
            generated feature names)
    :type dataset_identifier: str
    :param dataset_id: Identifier of the dataset in DataRobot AI Catalog
    :type dataset_id: str
    :param dataset_version_id: Identifier of the dataset version in DataRobot AI Catalog
    :type dataset_version_id: str, optional
    :param primary_temporal_key: Name of the column indicating time of record creation
    :type primary_temporal_key: str, optional
    :param feature_list_id: Specifies which feature list to use.
    :type feature_list_id: str, optional
    :param snapshot_policy: Policy to use  when creating a project or making predictions.
        If omitted, by default endpoint will use 'latest'.
        Must be one of the following values:
        'specified': Use specific snapshot specified by catalogVersionId
        'latest': Use latest snapshot from the same catalog item
        'dynamic': Get data from the source (only applicable for JDBC datasets)
    :type snapshot_policy: str, optional
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: Dataset definition
    :rtype: dict
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "dataset_identifier",
        "dataset_id",
        "dataset_version_id",
        "primary_temporal_key",
        "feature_list_id",
        "snapshot_policy",
    ]

    def __init__(
        self,
        *,
        dataset_identifier: str,
        dataset_id: Optional[str],
        dataset_version_id: str,
        snapshot_policy: str = SNAPSHOT_POLICY.LATEST,
        feature_list_id: Optional[str] = None,
        primary_temporal_key: Optional[str] = None,
        max_wait_sec: int = DATAROBOT_MAX_WAIT,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.dataset_identifier = dataset_identifier
        self.dataset_id = dataset_id
        self.dataset_version_id = dataset_version_id
        self.snapshot_policy = snapshot_policy
        self.feature_list_id = feature_list_id
        self.primary_temporal_key = primary_temporal_key
        self.max_wait_sec = max_wait_sec

    def execute(self, context: Context) -> dict:
        dataset_definition = dr.DatasetDefinition(
            identifier=self.dataset_identifier,
            catalog_id=self.dataset_id,
            catalog_version_id=self.dataset_version_id,
            feature_list_id=self.feature_list_id,
            primary_temporal_key=self.primary_temporal_key,
            snapshot_policy=self.snapshot_policy,  # requires a jdbc source
        )

        return dataset_definition.to_payload()


class DatasetRelationshipOperator(BaseDatarobotOperator):
    """
    Relationship between dataset defined in DatasetDefinition

    :param dataset1_identifier: Identifier of the first dataset in this relationship.
        This is specified in the identifier field of dataset_definition structure.
        If None, then the relationship is with the primary dataset.
    :type dataset1_identifier: str, optional
    :param dataset2_identifier: Identifier of the second dataset in this relationship.
        This is specified in the identifier field of dataset_definition schema.
    :type dataset2_identifier: str
    :param dataset1_keys: list of string (max length: 10 min length: 1)
        Column(s) from the first dataset which are used to join to the second dataset
    :type dataset1_keys: list
    :param dataset2_keys: list of string (max length: 10 min length: 1)
        Column(s) from the second dataset that are used to join to the first dataset
    :type dataset2_keys: list
    :param feature_derivation_window_start: How many time units of each dataset's
        primary temporal key into the past relative to the datetimePartitionColumn
        the feature derivation window should begin. Will be a negative integer,
        If present, the feature engineering Graph will perform time-aware joins.
    :type feature_derivation_window_start: int, optional
    :param feature_derivation_window_end: How many time units of each dataset's record
        primary temporal key into the past relative to the datetimePartitionColumn the
        feature derivation window should end.  Will be a non-positive integer, if present.
        If present, the feature engineering Graph will perform time-aware joins.
    :type feature_derivation_window_end: int, optional
    :param feature_derivation_window_time_unit: Time unit of the feature derivation window.
        One of ``datarobot.enums.AllowedTimeUnitsSAFER``
        If present, time-aware joins will be used.
        Only applicable when dataset1_identifier is not provided.
    :type feature_derivation_window_time_unit: int, optional
    :param feature_derivation_windows: List of feature derivation windows settings.
        If present, time-aware joins will be used.
        Only allowed when feature_derivation_window_start,
        feature_derivation_window_end and feature_derivation_window_time_unit are not provided.
    :type feature_derivation_windows: int, optional
    :param prediction_point_rounding: list of dict, or None
        Closest value of prediction_point_rounding_time_unit to round the prediction point
        into the past when applying the feature derivation window. Will be a positive integer,
        if present.Only applicable when dataset1_identifier is not provided.
    :type prediction_point_rounding: list[dict], optional
    :param prediction_point_rounding_time_unit: Time unit of the prediction point rounding.
        One of ``datarobot.enums.AllowedTimeUnitsSAFER`` Only applicable when dataset1_identifier
        is not provided.
    :type prediction_point_rounding_time_unit:  str, optional
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: Relationship definition
    :rtype: dict
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "dataset1_identifier",
        "dataset2_identifier",
        "dataset1_keys",
        "dataset2_keys",
        "feature_derivation_window_start",
        "feature_derivation_window_end",
        "feature_derivation_window_time_unit",
        "feature_derivation_windows",
        "prediction_point_rounding",
        "prediction_point_rounding_time_unit",
    ]

    def __init__(
        self,
        *,
        dataset2_identifier: str,
        dataset1_keys: list[str],
        dataset2_keys: list[str],
        dataset1_identifier: Optional[str] = None,
        feature_derivation_window_start: Optional[int] = None,
        feature_derivation_window_end: Optional[int] = None,
        feature_derivation_window_time_unit: Optional[int] = None,
        feature_derivation_windows: Optional[list[dict]] = None,
        prediction_point_rounding: Optional[int] = None,
        prediction_point_rounding_time_unit: Optional[str] = None,
        max_wait_sec: int = DATAROBOT_MAX_WAIT,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.dataset1_identifier = dataset1_identifier
        self.dataset2_identifier = dataset2_identifier
        self.dataset1_keys = dataset1_keys
        self.dataset2_keys = dataset2_keys
        self.feature_derivation_window_start = feature_derivation_window_start
        self.feature_derivation_window_end = feature_derivation_window_end
        self.feature_derivation_window_time_unit = feature_derivation_window_time_unit
        self.feature_derivation_windows = feature_derivation_windows
        self.prediction_point_rounding = prediction_point_rounding
        self.prediction_point_rounding_time_unit = prediction_point_rounding_time_unit
        self.max_wait_sec = max_wait_sec

    def execute(self, context: Context) -> dict:
        relationship = dr.Relationship(
            dataset1_identifier=self.dataset1_identifier,  # join profile
            dataset2_identifier=self.dataset2_identifier,  # to transactions
            dataset1_keys=self.dataset1_keys,  # on CustomerID
            dataset2_keys=self.dataset2_keys,
            feature_derivation_windows=self.feature_derivation_windows,
            prediction_point_rounding=self.prediction_point_rounding,
            prediction_point_rounding_time_unit=self.prediction_point_rounding_time_unit,
            feature_derivation_window_start=self.feature_derivation_window_start,
            feature_derivation_window_end=self.feature_derivation_window_end,
            feature_derivation_window_time_unit=self.feature_derivation_window_time_unit,
        )

        return relationship.to_payload()



# Copyright 2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
"""
Example of Aiflow DAG for DataRobot Batch Scoring using BigQuery as source and destination,
and using a preconfigured "DataRobot GCP Credentials" from Airflow Connection.
DataRobot GCP Credentials can be configured using Airflow UI (Admin->Connections) or Airflow API
Config example for this dag:
{
    "deployment_id": "put_your_deployment_id_here",
    "datarobot_gcp_credentials": "demo_bigquery_test_credentials",
    "score_settings": {
        "intake_settings": {
            "type": "bigquery",
            "dataset": "integration_demo",
            "table": "input_table_name",
            "bucket": "gcp_bucket_name",
        },
        "output_settings": {
            "type": "bigquery",
            "dataset": "integration_demo",
            "table": "output_table_name",
            "bucket": "gcp_bucket_name",
        },
        'passthrough_columns_set': 'all',
    },
}
"""

from datetime import datetime

from airflow.decorators import dag

from datarobot_provider.operators.credentials import GetOrCreateCredentialOperator
from datarobot_provider.operators.deployment import ScorePredictionsOperator
from datarobot_provider.sensors.datarobot import ScoringCompleteSensor


@dag(
    schedule=None,
    start_date=datetime(2023, 1, 1),
    tags=["example", "gcp", "bigquery"],
    params={
        "deployment_id": "put_your_deployment_id_here",
        "datarobot_gcp_credentials": "demo_bigquery_test_credentials",
        "score_settings": {
            "intake_settings": {
                "type": "bigquery",
                "dataset": "integration_demo",
                "table": "input_table_name",
                "bucket": "gcp_bucket_name",
            },
            "output_settings": {
                "type": "bigquery",
                "dataset": "integration_demo",
                "table": "output_table_name",
                "bucket": "gcp_bucket_name",
            },
            "passthrough_columns_set": "all",
        },
    },
)
def datarobot_bigquery_batch_scoring(deployment_id=None):
    if not deployment_id:
        raise ValueError("Invalid or missing `deployment_id` value")

    get_bigquery_credentials_op = GetOrCreateCredentialOperator(
        task_id="get_gcp_credentials",
        credentials_param_name="datarobot_gcp_credentials",
    )

    score_predictions_op = ScorePredictionsOperator(
        task_id="score_predictions",
        deployment_id=deployment_id,
        intake_credential_id=get_bigquery_credentials_op.output,
        output_credential_id=get_bigquery_credentials_op.output,
    )

    scoring_complete_sensor = ScoringCompleteSensor(
        task_id="check_scoring_complete",
        job_id=score_predictions_op.output,
    )

    (get_bigquery_credentials_op >> score_predictions_op >> scoring_complete_sensor)


datarobot_bigquery_batch_scoring_dag = datarobot_bigquery_batch_scoring()

# Staring from Airflow 2.5.1 Debug Executor is deprecated,
# dag.test() should be used for dag testing:
if __name__ == "__main__":
    datarobot_bigquery_batch_scoring_dag.test()



# Copyright 2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.

from typing import Union

from airflow.sensors.base import PokeReturnValue
from airflow.utils.context import Context
from datarobot import ModelJob

from datarobot_provider.sensors.model_insights import DataRobotJobSensor


class ModelTrainingJobSensor(DataRobotJobSensor):
    """
    Checks whether DataRobot Model Training Job is complete.

    :param project_id: DataRobot project ID
    :type project_id: str
    :param job_id: DataRobot Job ID
    :type job_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: False if not yet completed, PokeReturnValue(True, trained_model.id) if model training completed
    :rtype: bool | PokeReturnValue
    """

    def get_job_result(self, context: Context) -> Union[bool, PokeReturnValue]:
        trained_model = ModelJob.get_model(self.project_id, self.job_id)
        return PokeReturnValue(True, trained_model.id)



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any
from typing import Optional

from airflow.utils.context import Context
from datarobot import Credential

from datarobot_provider.hooks.credentials import CredentialsBaseHook
from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator


class GetOrCreateCredentialOperator(BaseDatarobotOperator):
    """
    Fetching credentials by Credential name and return Credentials ID.

    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: DataRobot Credentials ID
    :rtype: str
    """

    def __init__(
        self,
        *,
        credentials_param_name: str = "datarobot_credentials_name",
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.credentials_param_name = credentials_param_name

    def execute(self, context: Context) -> Optional[str]:
        credential_name = context["params"][self.credentials_param_name]
        # Trying to find a credential associated with provided credential name:
        for credential in Credential.list():
            if (
                credential.name == credential_name
                and credential.description != CredentialsBaseHook.default_credential_description
            ):
                self.log.info(
                    f"Found Credentials :{credential.name} , id={credential.credential_id} "
                    f"for param {self.credentials_param_name}"
                )
                return credential.credential_id
        else:
            # Trying to find an Airflow preconfigured credentials for provided credential name
            # to replicate credentials on DataRobot side:
            self.log.info(
                f"Credentials with name {credential_name} not found in DataRobot, trying to find "
                "Airflow connection with the same name"
            )
            hook = CredentialsBaseHook.get_hook(conn_id=credential_name)
            if hook.conn_type == "datarobot.datasource.jdbc":  # type: ignore[attr-defined]
                credentials, _, _ = hook.run()  # type: ignore[attr-defined]
            else:
                credentials, _ = hook.run()  # type: ignore[attr-defined]
            return credentials.credential_id



# Copyright 2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any
from typing import Tuple

from airflow.exceptions import AirflowException
from datarobot import Credential
from datarobot.models.data_store import DataStore
from datarobot.models.driver import DataDriver

from datarobot_provider.hooks.credentials import BasicCredentialsHook
from datarobot_provider.hooks.datarobot import DataRobotHook


class JDBCDataSourceHook(BasicCredentialsHook):
    """
    A hook that interacts with DataRobot via its public Python API library to
    manage JDBC connections with corresponding credentials.
    """

    conn_type = "datarobot.datasource.jdbc"
    hook_name = "DataRobot JDBC DataSource"

    @staticmethod
    def get_connection_form_widgets() -> dict[str, Any]:
        """Returns connection widgets to add to connection form."""
        from flask_appbuilder.fieldwidgets import BS3TextAreaFieldWidget
        from flask_appbuilder.fieldwidgets import BS3TextFieldWidget
        from flask_babel import lazy_gettext
        from wtforms import StringField

        return {
            "datarobot_connection": StringField(
                lazy_gettext("DataRobot Connection"),
                widget=BS3TextFieldWidget(),
                default="datarobot_default",
            ),
            "jdbc_driver": StringField(
                lazy_gettext("JDBC Driver"),
                widget=BS3TextFieldWidget(),
                default="",
            ),
            "jdbc_url": StringField(
                lazy_gettext("JDBC URL"),
                widget=BS3TextAreaFieldWidget(),
                default="jdbc:",
            ),
        }

    @staticmethod
    def get_ui_field_behaviour() -> dict:
        """Returns custom field behaviour."""
        return {
            "hidden_fields": ["host", "schema", "port", "extra"],
            "relabeling": {},
            "placeholders": {
                "datarobot_connection": "datarobot_default",
                "jdbc_driver": "",
                "jdbc_url": "jdbc:",
            },
        }

    def get_conn(self) -> tuple[Credential, dict, DataStore]:
        """Retrieving corresponding DataStore object or creating if not exist,
        updating it with new parameters in case of changes."""

        if not self.datarobot_credentials_conn_id:
            raise AirflowException("datarobot_credentials_conn_id is not defined")

        conn = self.get_connection(self.datarobot_credentials_conn_id)

        datarobot_connection_id = conn.extra_dejson.get("datarobot_connection", "")

        if not datarobot_connection_id:
            raise AirflowException("datarobot_connection is not defined")

        # Initialize DataRobot client by DataRobotHook
        DataRobotHook(datarobot_conn_id=datarobot_connection_id).run()

        jdbc_driver_name = conn.extra_dejson.get("jdbc_driver", "")

        if not jdbc_driver_name:
            raise AirflowException("jdbc_driver is not defined")

        jdbc_url = conn.extra_dejson.get("jdbc_url", "")

        if not jdbc_url:
            raise AirflowException("jdbc_url is not defined")

        if not conn.login:
            raise AirflowException("login is not defined")

        if not conn.password:
            raise AirflowException("password is not defined")

        credential = self.get_or_create_credential(conn)

        # For methods that accept credential data instead of credential ID
        credential_data = self.get_credential_data(conn)

        # Find the JDBC driver ID from name:
        for jdbc_drv_item in DataDriver.list():
            if jdbc_drv_item.canonical_name == jdbc_driver_name:
                jdbc_driver_id = jdbc_drv_item.id
                self.log.info(
                    f"Found JDBC Driver:{jdbc_drv_item.canonical_name} , id={jdbc_drv_item.id}"
                )
                break
        else:
            raise AirflowException(f'JDBC Driver "{jdbc_driver_name}" not found')

        # Check if DataStore created already:
        for data_store in DataStore.list():
            if data_store.canonical_name == self.datarobot_credentials_conn_id:
                self.log.info(
                    f"Found existing DataStore:{data_store.canonical_name} , id={data_store.id}"
                )
                break
        else:
            if not self.datarobot_credentials_conn_id:
                raise AirflowException("datarobot_credentials_conn_id is not defined")

            self.log.info(
                f"DataStore:{self.datarobot_credentials_conn_id} does not exist, trying to create it"
            )
            data_store = DataStore.create(
                data_store_type="jdbc",
                canonical_name=self.datarobot_credentials_conn_id,
                driver_id=jdbc_driver_id,
                jdbc_url=jdbc_url,
            )
            self.log.info(
                f"DataStore:{self.datarobot_credentials_conn_id} successfully created, id={data_store.id}"
            )

        if not data_store.params:
            raise AirflowException(f"DataStore:{self.datarobot_credentials_conn_id} has no params")

        if jdbc_url != data_store.params.jdbc_url or jdbc_driver_id != data_store.params.driver_id:
            self.log.info(
                f"Updating DataStore:{self.datarobot_credentials_conn_id} with new params..."
            )
            data_store.update(
                canonical_name=self.datarobot_credentials_conn_id,
                driver_id=jdbc_driver_id,
                jdbc_url=jdbc_url,
            )
            self.log.info(
                f"DataStore:{self.datarobot_credentials_conn_id} successfully updated, id={data_store.id}"
            )

        return credential, credential_data, data_store

    def run(self) -> Any:
        # get or create DataStore object updated with actual parameters
        return self.get_conn()

    def test_connection(self) -> Tuple[bool, str]:
        """Test DataRobot connection to JDBC DataSource"""
        try:
            _, credential_data, data_store = self.run()

            test_result = data_store.test(
                username=credential_data["user"], password=credential_data["password"]
            )

            success_massage = f"{test_result['message']}"
            return True, success_massage
        except Exception as e:
            return False, str(e)



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
"""
Config example for this dag:
{
    "datarobot_jdbc_connection": "datarobot_jdbc_test",
}
"""

from datetime import datetime

from airflow.decorators import dag

from datarobot_provider.operators.connections import GetOrCreateDataStoreOperator


@dag(
    schedule=None,
    start_date=datetime(2023, 1, 1),
    tags=["example"],
    # Default json config example:
    params={
        "datarobot_jdbc_connection": "datarobot_jdbc_test",
    },
)
def datarobot_test_datastore():
    dataset_connect_op = GetOrCreateDataStoreOperator(
        task_id="create_dataset_jdbc",
        connection_param_name="datarobot_jdbc_connection",
    )

    dataset_connect_op


datarobot_test_datastore_dag = datarobot_test_datastore()

if __name__ == "__main__":
    datarobot_test_datastore_dag.test()



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
"""
Config example for batch monitoring dag example:
{
    "deployment_id": "put_your_deployment_id_here",
    "datarobot_gcp_credentials": "demo_bigquery_test_credentials",
    "intake_settings": {
        "type": "bigquery",
        "dataset": "integration_demo",
        "table": "input_table_name",
        "bucket": "gcp_bucket_name",
    },
    "monitoring_columns": {
        "predictions_columns": [
            {
                "class_name": "True",
                "column_name": "target_True_PREDICTION"
            },
            {
                "class_name": "False",
                "column_name": "target_False_PREDICTION"
            }
        ]
        "association_id_column": "rowID",
        "actuals_value_column": "ACTUALS"
    }
}
"""

from datetime import datetime

from airflow.decorators import dag

from datarobot_provider.operators.credentials import GetOrCreateCredentialOperator
from datarobot_provider.operators.monitoring_job import BatchMonitoringOperator
from datarobot_provider.sensors.monitoring_job import MonitoringJobCompleteSensor


@dag(
    schedule=None,
    start_date=datetime(2023, 1, 1),
    tags=["example", "gcp", "bigquery", "monitoring"],
    params={
        "deployment_id": "put_your_deployment_id_here",
        "datarobot_gcp_credentials": "demo_bigquery_test_credentials",
        "monitoring_settings": {
            "intake_settings": {
                "type": "bigquery",
                "dataset": "integration_example_demo",
                "table": "actuals_demo",
                "bucket": "datarobot_demo_airflow",
            },
            "monitoring_columns": {
                "predictions_columns": [
                    {"class_name": "True", "column_name": "target_True_PREDICTION"},
                    {"class_name": "False", "column_name": "target_False_PREDICTION"},
                ],
                "association_id_column": "id",
                "actuals_value_column": "ACTUAL",
            },
        },
    },
)
def datarobot_batch_monitoring(deployment_id=None):
    if not deployment_id:
        raise ValueError("Invalid or missing `deployment_id` value")

    get_bigquery_credentials_op = GetOrCreateCredentialOperator(
        task_id="get_gcp_credentials",
        credentials_param_name="datarobot_gcp_credentials",
    )

    batch_monitoring_op = BatchMonitoringOperator(
        task_id="batch_monitoring",
        deployment_id=deployment_id,
        credential_id=get_bigquery_credentials_op.output,
    )

    batch_monitoring_complete_sensor = MonitoringJobCompleteSensor(
        task_id="check_monitoring_job_complete",
        job_id=batch_monitoring_op.output,
        poke_interval=15,  # status check each 15 sec
        mode="reschedule",
        timeout=7200,  # timeout after 2h (2*60*60sec = 7200 sec)
    )

    (get_bigquery_credentials_op >> batch_monitoring_op >> batch_monitoring_complete_sensor)


datarobot_batch_monitoring_dag = datarobot_batch_monitoring()

if __name__ == "__main__":
    datarobot_batch_monitoring_dag.test()



# Copyright 2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from collections.abc import Sequence
from typing import Any
from typing import Optional

import datarobot as dr
from airflow.utils.context import Context

from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator


class TrainModelOperator(BaseDatarobotOperator):
    """
    Submit a job to the queue to train a model from specific blueprint.
    :param project_id: DataRobot project ID
    :type project_id: str
    :param blueprint_id: DataRobot blueprint ID
    :type blueprint_id: str
    :param featurelist_id: The identifier of the featurelist to use.
        If not defined, the default for this project is used.
    :type featurelist_id: str, optional
    :source_project_id: Which project created this blueprint_id.
        If ``None``, it defaults to looking in this project.
        Note that you must have read permissions in this project.
    :source_project_id: str, optional
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: model training job ID
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "project_id",
        "blueprint_id",
        "featurelist_id",
        "source_project_id",
    ]

    def __init__(
        self,
        *,
        project_id: str,
        blueprint_id: str,
        featurelist_id: Optional[str] = None,
        source_project_id: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.blueprint_id = blueprint_id
        self.featurelist_id = featurelist_id
        self.source_project_id = source_project_id

    def validate(self) -> None:
        if not self.project_id:
            raise ValueError("project_id is required.")

        if not self.blueprint_id:
            raise ValueError("blueprint_id is required.")

    def execute(self, context: Context) -> str:
        project: dr.Project = dr.Project.get(self.project_id)
        blueprint = dr.Blueprint.get(self.project_id, self.blueprint_id)

        job_id = project.train(
            blueprint,
            sample_pct=context["params"].get("sample_pct", None),
            featurelist_id=self.featurelist_id,
            source_project_id=self.source_project_id,
            scoring_type=context["params"].get("scoring_type", None),
            training_row_count=context["params"].get("training_row_count", None),
            n_clusters=context["params"].get("n_clusters", None),
        )

        self.log.info(f"Model Training Job submitted job_id={job_id}")

        return job_id


class RetrainModelOperator(BaseDatarobotOperator):
    """
    Submit a job to the queue to retrain a model on a specific sample size and/or custom featurelist
    :param project_id: DataRobot project ID
    :type project_id: str
    :param model_id: DataRobot model ID
    :type model_id: str
    :param featurelist_id: The identifier of the featurelist to use.
        If not defined, the default for this project is used.
    :type featurelist_id: str, optional
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: model retraining job ID
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "project_id",
        "model_id",
        "featurelist_id",
    ]

    def __init__(
        self,
        *,
        project_id: str,
        model_id: str,
        featurelist_id: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.model_id = model_id
        self.featurelist_id = featurelist_id

    def validate(self) -> None:
        if self.project_id is None:
            raise ValueError("project_id is required.")

        if self.model_id is None:
            raise ValueError("model_id is required.")

    def execute(self, context: Context) -> str:
        model = dr.Model.get(self.project_id, self.model_id)

        job_id = model.train(
            featurelist_id=self.featurelist_id,
            sample_pct=context["params"].get("sample_pct", None),
            scoring_type=context["params"].get("scoring_type", None),
            training_row_count=context["params"].get("training_row_count", None),
        )

        self.log.info(f"Model Retraining Job submitted job_id={job_id}")

        return job_id



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from collections.abc import Sequence
from typing import Any
from typing import Optional

import datarobot as dr
from airflow.utils.context import Context

from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator

DATAROBOT_MAX_WAIT = 600


class StartAutopilotOperator(BaseDatarobotOperator):
    """
    Triggers DataRobot Autopilot to train set of models.

    :param project_id: DataRobot project ID
    :type project_id: str
    :param featurelist_id: Specifies which feature list to use.
    :type featurelist_id: str, optional
    :param relationships_configuration_id: ID of the relationships configuration to use.
    :type relationships_configuration_id: str, optional
    :param segmentation_task_id: The segmentation task that should be used to split the project
            for segmented modeling.
    :type segmentation_task_id: str, optional
    :param max_wait_sec: For some settings, an asynchronous task must be run to analyze the dataset.  max_wait
            governs the maximum time (in seconds) to wait before giving up.
    :type max_wait_sec: int, optional
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "project_id",
        "featurelist_id",
        "relationships_configuration_id",
        "segmentation_task_id",
    ]

    def __init__(
        self,
        *,
        project_id: str,
        featurelist_id: Optional[str] = None,
        relationships_configuration_id: Optional[str] = None,
        segmentation_task_id: Optional[str] = None,
        max_wait_sec: int = DATAROBOT_MAX_WAIT,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.featurelist_id = featurelist_id
        self.relationships_configuration_id = relationships_configuration_id
        self.segmentation_task_id = segmentation_task_id
        self.max_wait_sec = max_wait_sec

    def execute(self, context: Context) -> None:
        # Train models
        project = dr.Project.get(self.project_id)
        if project.target:
            raise ValueError(f"Models are already trained for project_id={project.id}")

        autopilot_settings = context["params"]["autopilot_settings"]

        self.log.info(
            f"Starting DataRobot Autopilot for project_id={project.id} "
            f"with settings={autopilot_settings}"
        )

        if self.featurelist_id:
            autopilot_settings["featurelist_id"] = self.featurelist_id

        if self.relationships_configuration_id:
            autopilot_settings["relationships_configuration_id"] = (
                self.relationships_configuration_id
            )

        if self.segmentation_task_id:
            autopilot_settings["segmentation_task_id"] = self.segmentation_task_id

        autopilot_settings["max_wait"] = self.max_wait_sec

        if (
            "datetime_partitioning_settings" in context["params"]
            and "partitioning_settings" in context["params"]
        ):
            raise ValueError(
                "parameters: datetime_partitioning_settings and partitioning_settings are mutually exclusive"
            )

        if "datetime_partitioning_settings" in context["params"]:
            project.set_datetime_partitioning(**context["params"]["datetime_partitioning_settings"])
        elif "partitioning_settings" in context["params"]:
            project.set_partitioning_method(**context["params"]["partitioning_settings"])

        if "advanced_options" in context["params"]:
            project.set_options(**context["params"]["advanced_options"])

        # finalize the project and start the autopilot
        project.analyze_and_model(**autopilot_settings)



# Copyright 2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.

import json
from typing import Any
from typing import Optional
from typing import Tuple
from typing import Union

from airflow.exceptions import AirflowException
from airflow.hooks.base import BaseHook
from airflow.models.connection import Connection
from datarobot.models.credential import Credential
from datarobot.models.data_store import DataStore

from datarobot_provider.hooks.datarobot import DataRobotHook


class CredentialsBaseHook(BaseHook):
    """
    Base class for DataRobot credential hooks that interacts with DataRobot
    via its public Python API library to manage stored credentials.

    :param datarobot_credentials_conn_id: Connection ID
    :type datarobot_credentials_conn_id: str
    """

    conn_name_attr = "datarobot_credentials_conn_id"
    hook_name = "DataRobot Credentials"
    default_credential_description = "Credentials managed by Airflow provider for Datarobot"

    def __init__(
        self,
        datarobot_credentials_conn_id: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.datarobot_credentials_conn_id = datarobot_credentials_conn_id

    def create_credentials(self, conn: Connection) -> Credential:
        """Creates DataRobot Credentials."""
        raise NotImplementedError()

    def get_credential_data(self, conn: Connection) -> dict:
        """Get credential data dict to use with methods that
        accept credential_data instead of credential ID"""
        raise NotImplementedError()

    def update_credentials(self, conn: Connection, credential: Credential) -> None:
        """Updates DataRobot Credentials"""
        raise NotImplementedError()

    def get_or_create_credential(self, conn: Connection) -> Credential:
        # Trying to find existing DataRobot Credentials:
        if not self.datarobot_credentials_conn_id:
            raise AirflowException("datarobot_credentials_conn_id is not defined")

        for credential in Credential.list():
            if credential.name == self.datarobot_credentials_conn_id:
                self.log.info(
                    f"Found Existing Credentials :{credential.name} , id={credential.credential_id}"
                )
                if (
                    credential.description
                    and self.default_credential_description in credential.description
                ):
                    self.log.info(
                        f"Trying to update provided credential:{credential.name} using Airflow preconfigured"
                        f" credentials"
                    )
                    self.update_credentials(conn, credential)
                    break
                else:
                    raise AirflowException(
                        f"Found Existing Credentials :{credential.name} , id={credential.credential_id}"
                        f" not managed by Airflow provider: {credential.description}"
                    )
        else:
            self.log.info(
                f"Credentials:{self.datarobot_credentials_conn_id} does not exist, trying to create it"
            )
            credential = self.create_credentials(conn)
            self.log.info(
                f"Credentials:{self.datarobot_credentials_conn_id} successfully created, id={credential.credential_id}"
            )
        return credential

    def get_conn(self) -> Union[tuple[Credential, dict, DataStore], tuple[Credential, dict]]:
        """Get or Create DataRobot associated credentials managed by Airflow provider."""
        if not self.datarobot_credentials_conn_id:
            raise AirflowException("datarobot_credentials_conn_id is not defined")

        conn = self.get_connection(self.datarobot_credentials_conn_id)

        datarobot_connection_id = conn.extra_dejson.get("datarobot_connection", "")

        if not datarobot_connection_id:
            raise AirflowException("datarobot_connection is not defined")

        # Initialize DataRobot client by DataRobotHook
        DataRobotHook(datarobot_conn_id=datarobot_connection_id).run()

        credential = self.get_or_create_credential(conn)
        credential_data = self.get_credential_data(conn)

        return credential, credential_data

    def run(self) -> Any:
        # get Credentials with credential_data
        return self.get_conn()

    def test_connection(self) -> Tuple[bool, str]:
        """Test that we can create DataRobot Credentials without errors"""
        try:
            credential, _ = self.run()
            credential = Credential.get(credential.credential_id)
            self.log.info(
                f"Test credential {credential.name} id={credential.credential_id} created"
            )
            # Airflow using randomly generated connection_id for to test connection,
            # so we need to delete it after creation:
            credential.delete()
            self.log.info(
                f"Test credential {credential.name} id={credential.credential_id} deleted"
            )
            return True, "Test creating DataRobot credentials success"
        except Exception as e:
            return False, str(e)


class BasicCredentialsHook(CredentialsBaseHook):
    hook_name = "DataRobot Basic Credentials"
    conn_type = "datarobot.credentials.basic"

    def __init__(
        self,
        datarobot_credentials_conn_id: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.datarobot_credentials_conn_id = datarobot_credentials_conn_id

    def create_credentials(self, conn: Connection) -> Credential:
        """Creates DataRobot Basic Credentials using provided login/password."""
        if not conn.login:
            raise AirflowException("login is not defined")

        if not conn.password:
            raise AirflowException("password is not defined")

        if not self.datarobot_credentials_conn_id:
            raise AirflowException("datarobot_credentials_conn_id is not defined")

        self.log.info(f"Creating Basic Credentials:{self.datarobot_credentials_conn_id}")
        credential = Credential.create_basic(
            name=self.datarobot_credentials_conn_id,
            user=conn.login,
            password=conn.password,
            description=self.default_credential_description,
        )

        return credential

    def update_credentials(self, conn: Connection, credential: Credential) -> None:
        """Updates DataRobot Basic Credentials using provided login/password."""
        basic_credentials = {
            "user": conn.login,
            "password": conn.password,
        }
        self.log.info(f"Updating Basic Credentials:{self.datarobot_credentials_conn_id}")

        try:
            credential.update(**basic_credentials)

        except Exception as e:
            self.log.error(
                f"Error updating Basic Credentials: {self.datarobot_credentials_conn_id}, message:{str(e)}"
            )
            raise AirflowException(
                f"Error updating Basic Credentials: {self.datarobot_credentials_conn_id}"
            ) from None

    def get_credential_data(self, conn: Connection) -> dict:
        # For methods that accept credential data instead of credential ID
        credential_data = {
            "credentialType": "basic",
            "user": conn.login,
            "password": conn.password,
        }
        return credential_data

    @staticmethod
    def get_connection_form_widgets() -> dict[str, Any]:
        """Returns connection widgets to add to connection form."""
        from flask_appbuilder.fieldwidgets import BS3TextFieldWidget
        from flask_babel import lazy_gettext
        from wtforms import StringField

        return {
            "datarobot_connection": StringField(
                lazy_gettext("DataRobot Connection"),
                widget=BS3TextFieldWidget(),
                default="datarobot_default",
            ),
        }

    @staticmethod
    def get_ui_field_behaviour() -> dict:
        """Returns custom field behaviour."""
        return {
            "hidden_fields": ["host", "schema", "port", "extra"],
            "relabeling": {},
            "placeholders": {
                "datarobot_connection": "datarobot_default",
                "login": "",
                "password": "",
            },
        }


class GoogleCloudCredentialsHook(CredentialsBaseHook):
    hook_name = "DataRobot GCP Credentials"
    conn_type = "datarobot.credentials.gcp"

    def parse_gcp_key(self, conn: Connection) -> dict:
        gcp_key = conn.extra_dejson.get("gcp_key", "")

        if not gcp_key:
            raise AirflowException("gcp_key is not defined")

        try:
            self.log.info("Trying to parse provided GCP key json")
            # removing newlines and parsing json:
            return json.loads(gcp_key.replace("\n", ""))

        except Exception as e:
            self.log.error(
                f"Error parsing GCP key json for credentials: {self.datarobot_credentials_conn_id}, message:{str(e)}"
            )
            raise AirflowException(
                f"Error parsing GCP key json for credentials: {self.datarobot_credentials_conn_id}"
            ) from None

    def create_credentials(self, conn: Connection) -> Credential:
        """Returns Google Cloud credentials for params in connection object"""
        gcp_key_json = self.parse_gcp_key(conn)

        if not self.datarobot_credentials_conn_id:
            raise AirflowException("datarobot_credentials_conn_id is not defined")

        self.log.info(f"Creating Google Cloud Credentials:{self.datarobot_credentials_conn_id}")
        credential = Credential.create_gcp(
            name=self.datarobot_credentials_conn_id,
            gcp_key=gcp_key_json,
            description=self.default_credential_description,
        )
        return credential

    def update_credentials(self, conn: Connection, credential: Credential) -> None:
        """Updates a Google Cloud credentials for params in connection object"""

        gcp_credentials = {
            "gcp_key": self.parse_gcp_key(conn),
        }

        if not self.datarobot_credentials_conn_id:
            raise AirflowException("datarobot_credentials_conn_id is not defined")

        self.log.info(f"Updating Google Cloud Credentials:{self.datarobot_credentials_conn_id}")
        try:
            credential.update(**gcp_credentials)  # type: ignore[arg-type]

        except Exception as e:
            self.log.error(
                f"Error updating Google Cloud Credentials: {self.datarobot_credentials_conn_id}, message:{str(e)}"
            )
            raise AirflowException(
                f"Error updating Google Cloud Credentials: {self.datarobot_credentials_conn_id}"
            ) from None

    def get_credential_data(self, conn: Connection) -> dict:
        # For methods that accept credential data instead of credential ID
        credential_data = {
            "credentialType": "gcp",
            "gcpKey": conn.extra_dejson.get("gcp_key", ""),
        }
        return credential_data

    @staticmethod
    def get_connection_form_widgets() -> dict[str, Any]:
        """Returns connection widgets to add to connection form."""
        from flask_appbuilder.fieldwidgets import BS3PasswordFieldWidget
        from flask_appbuilder.fieldwidgets import BS3TextFieldWidget
        from flask_babel import lazy_gettext
        from wtforms import StringField

        return {
            "datarobot_connection": StringField(
                lazy_gettext("DataRobot Connection"),
                widget=BS3TextFieldWidget(),
                default="datarobot_default",
            ),
            "gcp_key": StringField(
                lazy_gettext("GCP Key json content (Service Account)"),
                widget=BS3PasswordFieldWidget(),
            ),
        }

    @staticmethod
    def get_ui_field_behaviour() -> dict:
        """Returns custom field behaviour."""
        return {
            "hidden_fields": ["host", "schema", "port", "login", "password", "extra"],
            "relabeling": {},
            "placeholders": {
                "datarobot_connection": "datarobot_default",
                "gcp_key": "Enter a valid JSON string",
            },
        }


class AwsCredentialsHook(CredentialsBaseHook):
    hook_name = "DataRobot AWS Credentials"
    conn_type = "datarobot.credentials.aws"

    def create_credentials(self, conn: Connection) -> Credential:
        """Returns AWS credentials for params in connection object"""

        if not conn.login:
            raise AirflowException("aws_access_key_id is not defined")

        if not conn.password:
            raise AirflowException("aws_secret_access_key is not defined")

        # aws_session_token is optional:
        aws_session_token = conn.extra_dejson.get("aws_session_token", None)

        try:
            if not self.datarobot_credentials_conn_id:
                raise AirflowException("datarobot_credentials_conn_id is not defined")

            self.log.info(f"Creating AWS Credentials:{self.datarobot_credentials_conn_id}")
            credential = Credential.create_s3(
                name=self.datarobot_credentials_conn_id,
                aws_access_key_id=conn.login,
                aws_secret_access_key=conn.password,
                aws_session_token=aws_session_token,
                description=self.default_credential_description,
            )

            return credential

        except Exception as e:
            self.log.error(
                f"Error creating AWS Credentials: {self.datarobot_credentials_conn_id}, message:{str(e)}"
            )
            raise AirflowException(
                f"Error creating AWS Credentials: {self.datarobot_credentials_conn_id}"
            ) from None

    def update_credentials(self, conn: Connection, credential: Credential) -> None:
        """Updates AWS credentials for params in connection object"""

        if not conn.login:
            raise AirflowException("aws_access_key_id is not defined")

        if not conn.password:
            raise AirflowException("aws_secret_access_key is not defined")

        # aws_session_token is optional:
        aws_session_token = conn.extra_dejson.get("aws_session_token", None)

        try:
            self.log.info(f"Updating AWS Credentials:{self.datarobot_credentials_conn_id}")

            aws_credentials = {
                "aws_access_key_id": conn.login,
                "aws_secret_access_key": conn.password,
                "aws_session_token": aws_session_token,
            }

            credential.update(**aws_credentials)

        except Exception as e:
            self.log.error(
                f"Error updating AWS Credentials: {self.datarobot_credentials_conn_id}, message:{str(e)}"
            )
            raise AirflowException(
                f"Error updating AWS Credentials: {self.datarobot_credentials_conn_id}"
            ) from None

    def get_credential_data(self, conn: Connection) -> dict:
        # For methods that accept credential data instead of credential ID
        credential_data = {
            "credentialType": "s3",
            "awsAccessKeyId": conn.login,
            "awsSecretAccessKey": conn.password,
        }
        aws_session_token = conn.extra_dejson.get("aws_session_token", "")

        if aws_session_token:
            # if AWS Session Token is not empty:
            credential_data["awsSessionToken"] = aws_session_token
        return credential_data

    @staticmethod
    def get_connection_form_widgets() -> dict[str, Any]:
        """Returns connection widgets to add to connection form."""
        from flask_appbuilder.fieldwidgets import BS3TextAreaFieldWidget
        from flask_appbuilder.fieldwidgets import BS3TextFieldWidget
        from flask_babel import lazy_gettext
        from wtforms import StringField

        return {
            "datarobot_connection": StringField(
                lazy_gettext("DataRobot Connection"),
                widget=BS3TextFieldWidget(),
                default="datarobot_default",
            ),
            "aws_session_token": StringField(
                lazy_gettext("AWS session token"),
                widget=BS3TextAreaFieldWidget(),
            ),
        }

    @staticmethod
    def get_ui_field_behaviour() -> dict:
        """Returns custom field behaviour."""
        return {
            "hidden_fields": ["host", "schema", "port", "extra"],
            "relabeling": {
                "login": "AWS Access Key ID",
                "password": "AWS Secret Access Key",
            },
            "placeholders": {"datarobot_connection": "datarobot_default"},
        }


class AzureStorageCredentialsHook(CredentialsBaseHook):
    hook_name = "DataRobot Azure Storage Credentials"
    conn_type = "datarobot.credentials.azure"

    def create_azure_connection_string(self, conn: Connection) -> str:
        if not conn.login:
            raise AirflowException("Storage Account Name is not defined")

        if not conn.password:
            raise AirflowException("Storage Account Key is not defined")

        azure_connection_string = "DefaultEndpointsProtocol=https;"
        "AccountName={};AccountKey={};EndpointSuffix=core.windows.net".format(
            conn.login, conn.password
        )
        return azure_connection_string

    def create_credentials(self, conn: Connection) -> Credential:
        """Returns Azure Storage credentials for params in connection object"""

        try:
            if not self.datarobot_credentials_conn_id:
                raise AirflowException("datarobot_credentials_conn_id is not defined")

            self.log.info(
                f"Creating Azure Storage Credentials: {self.datarobot_credentials_conn_id}"
            )
            credential = Credential.create_azure(
                name=self.datarobot_credentials_conn_id,
                azure_connection_string=self.create_azure_connection_string(conn),
                description=self.default_credential_description,
            )

            return credential

        except Exception as e:
            self.log.error(
                f"Error creating Azure Credentials: {self.datarobot_credentials_conn_id}, message:{e}"
            )
            raise AirflowException(
                f"Error creating Azure Credentials: {self.datarobot_credentials_conn_id}"
            ) from None

    def update_credentials(self, conn: Connection, credential: Credential) -> None:
        """Updates Azure Storage credentials for params in connection object"""

        azure_credentials = {"azure_connection_string": self.create_azure_connection_string(conn)}
        self.log.info(f"Updating Azure Storage Credentials: {self.datarobot_credentials_conn_id}")
        try:
            credential.update(**azure_credentials)

        except Exception as e:
            self.log.error(
                f"Error updating Azure Storage Credentials: {self.datarobot_credentials_conn_id}, message:{str(e)}"
            )
            raise AirflowException(
                f"Error updating Azure Storage Credentials: {self.datarobot_credentials_conn_id}"
            ) from None

    def get_credential_data(self, conn: Connection) -> dict:
        # For methods that accept credential data instead of credential ID
        credential_data = {
            "credentialType": "azure",
            "azureConnectionString": "DefaultEndpointsProtocol=https;"
            "AccountName={};AccountKey={};EndpointSuffix=core.windows.net".format(
                conn.login, conn.password
            ),
        }
        return credential_data

    @staticmethod
    def get_connection_form_widgets() -> dict[str, Any]:
        """Returns connection widgets to add to connection form."""
        from flask_appbuilder.fieldwidgets import BS3TextFieldWidget
        from flask_babel import lazy_gettext
        from wtforms import StringField

        return {
            "datarobot_connection": StringField(
                lazy_gettext("DataRobot Connection"),
                widget=BS3TextFieldWidget(),
                default="datarobot_default",
            ),
        }

    @staticmethod
    def get_ui_field_behaviour() -> dict:
        """Returns custom field behaviour."""
        return {
            "hidden_fields": ["host", "schema", "port", "extra"],
            "relabeling": {
                "login": "Azure Storage Account Name",
                "password": "Azure Storage Account Key",
            },
            "placeholders": {"datarobot_connection": "datarobot_default"},
        }


class OAuthCredentialsHook(CredentialsBaseHook):
    hook_name = "DataRobot OAuth Credentials"
    conn_type = "datarobot.credentials.oauth"

    def create_credentials(self, conn: Connection) -> Credential:
        """Returns OAuth credentials for params in connection object"""

        if not conn.login:
            raise AirflowException("OAuth Client Id is not defined")

        if not conn.password:
            raise AirflowException("OAuth Token is not defined")

        refresh_token = conn.extra_dejson.get("refresh_token", "")
        if not refresh_token:
            raise AirflowException("OAuth Refresh Token is not defined")

        try:
            if not self.datarobot_credentials_conn_id:
                raise AirflowException("datarobot_credentials_conn_id is not defined")

            self.log.info(f"Creating OAuth Credentials: {self.datarobot_credentials_conn_id}")
            credential = Credential.create_oauth(
                name=self.datarobot_credentials_conn_id,
                token=conn.password,
                refresh_token=refresh_token,
                description=self.default_credential_description,
            )

            return credential

        except Exception as e:
            self.log.error(
                f"Error creating OAuth Credentials: {self.datarobot_credentials_conn_id}, message:{e}"
            )
            raise AirflowException(
                f"Error creating OAuth Credentials: {self.datarobot_credentials_conn_id}"
            ) from None

    def update_credentials(self, conn: Connection, credential: Credential) -> None:
        """Updates OAuth credentials for params in connection object"""

        if not conn.password:
            raise AirflowException("OAuth Token is not defined")

        refresh_token = conn.extra_dejson.get("refresh_token", "")
        if not refresh_token:
            raise AirflowException("OAuth Refresh Token is not defined")

        oauth_credentials = {
            "token": conn.password,
            "refresh_token": refresh_token,
        }

        self.log.info(f"Updating OAuth Credentials: {self.datarobot_credentials_conn_id}")
        try:
            credential.update(**oauth_credentials)

        except Exception as e:
            self.log.error(
                f"Error updating OAuth Credentials: {self.datarobot_credentials_conn_id}, message:{str(e)}"
            )
            raise AirflowException(
                f"Error updating OAuth Credentials: {self.datarobot_credentials_conn_id}"
            ) from None

    def get_credential_data(self, conn: Connection) -> dict:
        # For methods that accept credential data instead of credential ID
        credential_data = {
            "credentialType": "oauth",
            "oauthRefreshToken": conn.extra_dejson.get("refresh_token", ""),
            "oauthClientId": conn.login,
            "oauthClientSecret": conn.password,
        }
        return credential_data

    @staticmethod
    def get_connection_form_widgets() -> dict[str, Any]:
        """Returns connection widgets to add to connection form."""
        from flask_appbuilder.fieldwidgets import BS3PasswordFieldWidget
        from flask_appbuilder.fieldwidgets import BS3TextFieldWidget
        from flask_babel import lazy_gettext
        from wtforms import StringField

        return {
            "refresh_token": StringField(
                lazy_gettext("OAuth Refresh Token"),
                widget=BS3PasswordFieldWidget(),
            ),
            "datarobot_connection": StringField(
                lazy_gettext("DataRobot Connection"),
                widget=BS3TextFieldWidget(),
                default="datarobot_default",
            ),
        }

    @staticmethod
    def get_ui_field_behaviour() -> dict:
        """Returns custom field behaviour."""
        return {
            "hidden_fields": ["host", "schema", "port", "extra"],
            "relabeling": {
                "login": "OAuth Client Id",
                "password": "OAuth Token",
            },
            "placeholders": {"datarobot_connection": "datarobot_default"},
        }



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
"""
Config example for this dag:
{
    "datarobot_jdbc_connection": "datarobot_jdbc_test",
    "dataset_name": "test_dataset_name",
    "table_schema": "integration_demo",
    "table_name": "test_table",
    "query": 'SELECT * FROM "integration_demo"."test_table"',
    "persist_data_after_ingestion": False,
    "do_snapshot": False,
}
"""

from datetime import datetime

from airflow.decorators import dag

from datarobot_provider.operators.ai_catalog import CreateDatasetFromDataStoreOperator
from datarobot_provider.operators.credentials import GetOrCreateCredentialOperator
from datarobot_provider.operators.datarobot import CreateProjectOperator


@dag(
    schedule=None,
    start_date=datetime(2023, 1, 1),
    tags=["example"],
    # Default json config example:
    params={
        "project_name": "test_project_name",
        "datarobot_jdbc_connection": "datarobot_jdbc_test_connection_name",
        "dataset_name": "test_jdbc_dataset_name",
        "table_schema": "test_jdbc_table_schema",
        "table_name": "test_jdbc_table_name",
        "persist_data_after_ingestion": False,
        "do_snapshot": False,
    },
)
def datarobot_dynamic_jdbc_dataset():
    dataset_connect_op = CreateDatasetFromDataStoreOperator(
        task_id="create_dataset_jdbc",
    )

    # In case of dynamic dataset we should provide credential_id from connection
    get_jdbc_credentials_op = GetOrCreateCredentialOperator(
        task_id="get_jdbc_credentials",
        credentials_param_name="datarobot_jdbc_connection",
    )

    create_project_op = CreateProjectOperator(
        task_id="create_project",
        dataset_id=dataset_connect_op.output,
        # In case of dynamic dataset we should provide credential_id
        credential_id=get_jdbc_credentials_op.output,
    )

    dataset_connect_op >> create_project_op


datarobot_dynamic_jdbc_dataset_dag = datarobot_dynamic_jdbc_dataset()

if __name__ == "__main__":
    datarobot_dynamic_jdbc_dataset_dag.test()



# Copyright 2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from collections.abc import Sequence
from typing import Any
from typing import Optional

import datarobot as dr
from airflow.exceptions import AirflowFailException
from airflow.utils.context import Context
from datarobot.insights import ShapPreview

from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator


class ComputeFeatureImpactOperator(BaseDatarobotOperator):
    """
    Creates Feature Impact job in DataRobot.
    :param project_id: DataRobot project ID
    :type project_id: str
    :param model_id: DataRobot model ID
    :type model_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: Feature Impact job ID
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "project_id",
        "model_id",
    ]

    def __init__(
        self,
        *,
        project_id: str,
        model_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.model_id = model_id

    def validate(self) -> None:
        if not self.project_id:
            raise ValueError("project_id is required to compute Feature Impact.")

        if not self.model_id:
            raise ValueError("model_id is required to compute Feature Impact.")

    def execute(self, context: Context) -> str:
        model = dr.models.Model.get(self.project_id, self.model_id)

        job = model.request_feature_impact()

        self.log.info(f"Feature Impact Job submitted, job_id={job.id}")

        return job.id


class ComputeFeatureEffectsOperator(BaseDatarobotOperator):
    """
    Submit request to compute Feature Effects for the model.
    :param project_id: DataRobot project ID
    :type project_id: str
    :param model_id: DataRobot model ID
    :type model_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: Feature Effects job ID
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "project_id",
        "model_id",
    ]

    def __init__(
        self,
        *,
        project_id: str,
        model_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.model_id = model_id

    def validate(self) -> None:
        if not self.project_id:
            raise ValueError("project_id is required to compute Feature Effects.")

        if not self.model_id:
            raise ValueError("model_id is required to compute Feature Effects.")

    def execute(self, context: Context) -> str:
        model = dr.models.Model.get(self.project_id, self.model_id)

        job = model.request_feature_effect()

        self.log.info(f"Feature Effects Job submitted, job_id={job.id}")

        return job.id


class ComputeShapPreviewOperator(BaseDatarobotOperator):
    """
    Creates SHAP preview job in DataRobot.

    Parameters
    ----------
    project_id : str
        DataRobot project ID
    model_id : str
        DataRobot model ID
    datarobot_conn_id : str, optional
        Connection ID, defaults to `datarobot_default`

    Returns
    -------
    str
        SHAP preview result id
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "model_id",
    ]

    def __init__(
        self,
        *,
        model_id: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.model_id = str(model_id) if model_id else None

    def validate(self) -> None:
        if not self.model_id:
            raise AirflowFailException("The `model_id` parameter is required.")

    def execute(self, context: Context) -> str:
        shap_preview = ShapPreview.create(entity_id=self.model_id)  # type: ignore[arg-type]

        return shap_preview.id



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
"""
Config example for this dag:
{
    "datarobot_jdbc_connection": "datarobot_jdbc_demo",
    "dataset_name": "integration_example_demo",
    "table_schema": "integration_example_demo",
    "table_name": "actuals_demo",
}
"""

from datetime import datetime

from airflow.decorators import dag

from datarobot_provider.operators.ai_catalog import CreateDatasetVersionOperator
from datarobot_provider.operators.ai_catalog import CreateOrUpdateDataSourceOperator
from datarobot_provider.operators.connections import GetOrCreateDataStoreOperator
from datarobot_provider.operators.credentials import GetOrCreateCredentialOperator


@dag(
    schedule=None,
    start_date=datetime(2023, 1, 1),
    tags=["example", "mlops"],
    # Default json config example:
    params={
        "datarobot_jdbc_connection": "datarobot_jdbc_demo",
        "dataset_name": "integration_example_demo",
        "table_schema": "integration_example_demo",
        "table_name": "actuals_demo",
    },
)
def datarobot_dataset_new_version(dataset_id=None):
    get_credentials_op = GetOrCreateCredentialOperator(
        task_id="get_jdbc_credentials",
        credentials_param_name="datarobot_jdbc_connection",
    )

    get_datastore_op = GetOrCreateDataStoreOperator(
        task_id="get_datastore",
        connection_param_name="datarobot_jdbc_connection",
    )

    get_datasource_op = CreateOrUpdateDataSourceOperator(
        task_id="get_datasource",
        data_store_id=get_datastore_op.output,
    )

    create_dataset_version_op = CreateDatasetVersionOperator(
        task_id="create_new_version_of_dataset",
        dataset_id=dataset_id,
        datasource_id=get_datasource_op.output,
        credential_id=get_credentials_op.output,
    )

    get_datastore_op >> get_datasource_op >> get_credentials_op >> create_dataset_version_op


datarobot_dataset_new_version_dag = datarobot_dataset_new_version()

if __name__ == "__main__":
    datarobot_dataset_new_version_dag.test()



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from datetime import datetime

from airflow.decorators import dag
from airflow.decorators import task
from datarobot.enums import MODEL_REPLACEMENT_REASON

from datarobot_provider.operators.deployment import GetDeploymentModelOperator
from datarobot_provider.operators.deployment import ReplaceModelOperator


@dag(
    schedule=None,
    start_date=datetime(2023, 1, 1),
    tags=["example", "mlops"],
)
def deployment_replace_model(deployment_id=None, new_model_id=None):
    if not deployment_id:
        raise ValueError("Invalid or missing `deployment_id` value")
    if not new_model_id:
        raise ValueError("Invalid or missing `new_model_id` value")

    get_deployment_model_before_op = GetDeploymentModelOperator(
        task_id="get_deployment_model_before",
        deployment_id=deployment_id,
    )

    replace_deployment_model_op = ReplaceModelOperator(
        task_id="replace_deployment_model",
        deployment_id=deployment_id,
        new_model_id=new_model_id,
        reason=MODEL_REPLACEMENT_REASON.ACCURACY,
    )

    get_deployment_model_after_op = GetDeploymentModelOperator(
        task_id="get_deployment_model_after",
        deployment_id=deployment_id,
    )

    @task(task_id="example_code_python")
    def deployment_model_check_example(deployment_model_before, deployment_model_after):
        """Example of custom logic based on comparing old and replaced from the deployment."""

        # Put your custom logic here:
        return deployment_model_before["id"] != deployment_model_after["id"]

    example_deployment_model_check = deployment_model_check_example(
        deployment_model_before=get_deployment_model_before_op.output,
        deployment_model_after=get_deployment_model_after_op.output,
    )

    (
        get_deployment_model_before_op
        >> replace_deployment_model_op
        >> get_deployment_model_after_op
        >> example_deployment_model_check
    )


deployment_replace_model_dag = deployment_replace_model()

if __name__ == "__main__":
    deployment_replace_model_dag.test()
